#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <wait.h>
#include <pthread.h>
#include <string.h>

#define bool int
#define true 1
#define false 0
#define print(...) do{printf(__VA_ARGS__); fflush(stdout); } while(0)

#define FOR_MONITOR

typedef struct node_struct node;

struct node_struct {
    pthread_mutex_t *node_lock;
    node* next;
};

node *new_node() {
    node *n = (node*) malloc(sizeof(node));
    n->node_lock = malloc(sizeof(pthread_mutex_t));
    pthread_mutex_init(n->node_lock, NULL);
    n->next = NULL; //bonne pratique!
    return n;
}

void free_node(node *n) {
    pthread_mutex_destroy(n->node_lock);
    free(n);
}

typedef struct {
    pthread_mutex_t *condition_lock;
    int in_use;
    int id;
    node *n;
} condition;

typedef struct {
    condition **conditions;
    condition *c;   //speciale, fake
} monitor;

condition *new_condition() {
    condition *c = malloc(sizeof(condition));
    c->condition_lock = malloc(sizeof(pthread_mutex_t));

    pthread_mutex_init(c->condition_lock, NULL);
    c->in_use = false;
    c->n = NULL;

    return c;
}

void shorten_condition_list(condition *c) {
    if(c->n == NULL) {
        print("\tI concur my dear, no need to unlock, it is indeed null\n");
        return;
    }

    print("\tUnlocking head node\n");
    node *next = c->n->next;

    pthread_mutex_unlock(c->n->node_lock); //big
    pthread_mutex_unlock(c->n->node_lock); //dumb

    free_node(c->n);
    c->n = next;
}

void depth(condition *c) {
    int i = 0;
    for(node *n=c->n; n != NULL; n=n->next) i++;
    print("\t\tCondition waitlist depth: %d\n", i);
}

void extend_condition(condition *c) {
    depth(c);
    node *n = new_node();

    if(c->n == NULL) c->n = n;
    else {
        node *old_node = c->n;
        while(1) {
            if(old_node->next == NULL) break;
            old_node = old_node->next;
        }
        old_node->next = n;
    }

    pthread_mutex_lock(n->node_lock);   //marche
    pthread_mutex_unlock(c->condition_lock); //laisse d'autres tester la condition
    print("\tMe sleepy\n");
    pthread_mutex_lock(n->node_lock);   //s'auto bloque

    print("\tME WAKEY!\n");
    //une fois signalés

    pthread_mutex_lock(c->condition_lock);
    c->in_use = true;
    pthread_mutex_unlock(c->condition_lock);
}

void enter_monitor(monitor *m);
void leave_monitor(monitor *m);

void wait_condition(condition *c, monitor *m) {
    print("\tGrabbing that tasty condition\n");
    pthread_mutex_lock(c->condition_lock);

    if(c->in_use) {
        if(m != NULL) leave_monitor(m);
        extend_condition(c);
        if(m != NULL) enter_monitor(m);
    } else {
        c->in_use = true;
        pthread_mutex_unlock(c->condition_lock);
    }

    depth(c);
}

void _signal_condition(condition *c) {
    pthread_mutex_lock(c->condition_lock);
    c->in_use = false;
    shorten_condition_list(c);  //wakey wakey
    pthread_mutex_unlock(c->condition_lock);
    print("\tDone with the condition, it was overrated IMO\n");
}

void signal_and_wait(condition *c, monitor *m) {
    _signal_condition(c);    //plus besoin de la ressources
    leave_monitor(m);           //mais on doit sortir pour laisser d'autres entrer
    enter_monitor(m);           //mais on est encore dans le monitor, donc on doit rentrer après!
}

void signal_and_continue(condition *c, monitor *m) {    //TODO
    _signal_condition(c);    //plus besoin de la ressources
    leave_monitor(m);           //mais on doit sortir pour laisser d'autres entrer
    enter_monitor(m);           //mais on est encore dans le monitor, donc on doit rentrer après!
}

monitor *new_monitor(condition **conditions) {
    monitor *m = malloc(sizeof(monitor));

    m->c = new_condition();
    m->conditions = conditions;

    return m;
}

void enter_monitor(monitor *m) {    //locks on fake condition
    wait_condition(m->c, NULL);
}

void leave_monitor(monitor *m) {    //releases threads stuck in fake condition
    _signal_condition(m->c);
}

int counter = 0;
monitor *m;
condition *c;
void *printy(void *args) {
    enter_monitor(m);
    wait_condition(c, m);
    //sleep(5);
    //wait_condition(c, m);
    for(int i=0; i<10; i++) {
        print("COUNTER: %d\n", counter);
        counter++;
    }
    signal_and_wait(c, m);
    leave_monitor(m);
    print("\tDONE!\n");
    pthread_exit(0);
}

int main() {
    m = new_monitor(NULL);
    c = new_condition();

    int nb_t = 10;

    pthread_t threads[nb_t];
    for(int i=0; i<nb_t; i++) {
        pthread_create(&threads[i], NULL, printy, NULL);
    }

    for(int i=0; i<nb_t; i++) {
        pthread_join(threads[i], NULL);
    }
}
